#include <stdbool.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <errno.h>
#include <limits.h>
#include <sys/wait.h>
#include <stdarg.h>
#include <sys/eventfd.h>
#include <time.h>

#include "mctd.h"
#include "mctd-startup-parser.h"
#include "mctd-log.h"
#include "mctd-mux.h"
#include "mctd-serial.h"
#include "mctd-atcmd.h"
#include "mctd-wvdial-conf.h"
#include "mctd-modem-control.h"

#define URC_SYSSTART "^SYSSTART"
#define URC_SHUTDOWN "^SHUTDOWN"

#define AT_RESPONSE_OK "\r\n\r\nOK\r\n"
#define AT_RESPONSE_OK_LEN 8

#define EMS31_PWR_CTL_PATH "/sys/devices/soc0/ems31-reset/ems31_power_ctrl"
#define EMS31_GPIO25_PATH "/sys/class/gpio/gpio90/value"

static char *_operator_id = NULL;
static bool _auto_dial = false;
static bool _no_sim_card = false;
static bool _checked_sim_card = false;
static bool _is_suspend = false;
static bool _ping_check = false;
static int _check_register_interval_sec = 0;

int _multiplex;
static char *_manufacturer = NULL;
static char *_model = NULL;
static char *_revision = NULL;
static char *_app_revision = NULL;
static char *_phone_number = NULL;
static char *_imei = NULL;
static int _rsrq = -1;
static int _rsrp = -1;
static int _sim_slot = 0;
static char *_registerd_operator_id = NULL;
static int _ping_failed_threshold = 1;
static char *_ping_ip_addr = NULL;
static int _failed_ping_check_count = 0;
static int _network_restart_count = 0;

static bool check_param(const char *param, char *expect_str) {
    char *s;
    bool ret;
    int len = strlen(expect_str) + 2;

    s = malloc(len + 1);
    snprintf(s, len + 1, "\"%s\"", expect_str);

    if (strcmp(param, expect_str) == 0 ||
    strcmp(param, s) == 0) {
        ret = true;
    } else {
        ret = false;
    }

    free(s);
    return ret;
}

static void reap_defunct_child_proc()
{
    pid_t pid;
    int status;
    do {
        pid = waitpid(-1, &status, WNOHANG);
        if (pid < 0) {
            if (errno == ECHILD)
                break;

            if (errno != EINTR) {
                perror("waitpid");
                break;
            }
        }
    } while (pid != 0);
}

static int bin_execution(bool background, const char *arg, ...)
{
    int argc;
    pid_t pid, n;
    int status;

    va_list ap;
    va_start(ap, arg);
    for (argc = 1; va_arg(ap, const char *); argc++) {
        if (argc == INT_MAX) {
            va_end(ap);
            return -1;
        }
    }
    va_end(ap);

    char *argv[argc + 1];
    va_start(ap, arg);
    argv[0] = (char *) arg;

    for (int i = 1; i <= argc; i++) {
        argv[i] = va_arg(ap, char *);
    }
    va_end(ap);

    pid = fork();

    if(pid == -1) {
        perror("fork");
        status = -1;
    }

    if(pid == 0) {
        execvp(argv[0], argv);
        log_err("command execution failed.");
        exit(1);
    } else {
        if (!background) {
            do {
                n = waitpid(pid, &status, 0);
                if (n < 0) {
                    if (errno == ECHILD) {
                        status = 0;
                        break;
                    }

                    if (errno != EINTR) {
                        perror("waitpid");
                        status = -1;
                        break;
                    }
                }
            } while (n != pid);
        } else {
            status = 0;
        }
    }
    return status;
}

static bool is_wvdial_running() {
    reap_defunct_child_proc();
    if (bin_execution(false, "pgrep", "wvdial", NULL) == 0) {
        return true;
    }
    return false;
}

static void set_errot_format_with_fd(int fd)
{
    char *response = NULL;
    int result;

    result = mctd_atcmd_send(fd, "AT+CMEE?", "+CMEE: 2", &response, NULL);
    if (result != MCTD_RESULT_OK) {
        log_info("AT+CMEE? failed");
        return;
    }

    if (response) {
        /* already set 2 */
        free(response);
        return;
    }

    mctd_atcmd_send(fd, "AT+CMEE=2", NULL, NULL, NULL);
}

static int set_serial_baudrate()
{
    int fd;
    int timeout;

    /* open serial with baudrate 460800 */
    fd = mctd_serial_open();
    if (fd < 0) {
        log_warn("open serial failed.");
        return mctd_serial_convert_error(fd);
    }

    timeout = 3;
    if (mctd_atcmd_send(fd, "AT", NULL, NULL, &timeout) == MCTD_RESULT_OK) {
        mctd_serial_close(fd);
        return MCTD_RESULT_OK;
    }

    mctd_serial_close(fd);

    /* open serial with baudrate 115200 */
    fd = mctd_serial_open_baudrate_115200();
    if (fd < 0)
        return mctd_serial_convert_error(fd);

    timeout = 3;
    if (mctd_atcmd_send(fd, "AT", NULL, NULL, &timeout) != MCTD_RESULT_OK) {
        /* baurate is unknown */
        mctd_serial_close(fd);
        return MCTD_RESULT_ERR_SERIAL;
    }

    /* set baudrate 460800 */
    if (mctd_atcmd_send(fd, "AT+IPR=460800", NULL, NULL, NULL) != MCTD_RESULT_OK) {
        mctd_serial_close(fd);
        return MCTD_RESULT_ERR_SERIAL;
    }

    mctd_serial_close(fd);

    /* open serial with baudrate 460800 */
    fd = mctd_serial_open();
    if (fd < 0) {
        log_warn("mctd_serial_open failed.");
        return mctd_serial_convert_error(fd);
    }

    timeout = 3;
    if (mctd_atcmd_send(fd, "AT", NULL, NULL, NULL) == MCTD_RESULT_OK) {
        mctd_serial_close(fd);
        return MCTD_RESULT_OK;
    } else {
        log_warn("send AT failed, on baudrate 4608000.");
        mctd_serial_close(fd);
        return MCTD_RESULT_ERR_SERIAL;
    }
}

static bool is_calling()
{
    if (_multiplex == MUX_STATUS_DISABLE) {
        return is_wvdial_running();
    }
    return false;
}

static void cutCRandLF(char *p)
{
    if (!p)
        return;

    while(1) {
        if ((p[strlen(p) -1] == '\r') || (p[strlen(p) -1] == '\n'))
            p[strlen(p) -1] = '\0';
        else
            break;
    }
}

static int iccid_with_fd(int fd, char **iccid)
{
    int result;
    char *response = NULL;
    char *current = NULL;
    int iccid_len;

    result = mctd_atcmd_send(fd, "AT+CCID?", "+CCID:", &response, NULL);
    if (result != MCTD_RESULT_OK)
        return result;

    if (!response)
        return MCTD_RESULT_ERR_PROCESS;

    current = strtok(response, ",");
    if (!current) {
        result = MCTD_RESULT_ERR_NO_DATA;
        goto bail;
    }

    if (strlen(current) <= strlen("+CCID: ")) {
        result = MCTD_RESULT_ERR_NO_DATA;
        goto bail;
    }

    if (strcmp(&current[strlen("+CCID: ")], "\"\"") == 0) {
        result = MCTD_RESULT_ERR_NO_DATA;
        goto bail;
    }

    iccid_len = strlen(current) - strlen("+CCID: ");
    *iccid = malloc(iccid_len + 1);
    if (!*iccid) {
        perror("malloc");
        log_err("malloc failed");
        result = MCTD_RESULT_ERR_MEMORY;
        goto bail;
    }
    strcpy(*iccid, &current[strlen("+CCID: ")]);

bail:
    free(response);
    return result;
}

static bool has_sim_card_with_fd(int fd)
{
    char *iccid = NULL;
    int i;

    if (_no_sim_card)
        return false;

    if (_checked_sim_card)
        return true;

    for (i = 0; i < 3 ; i++) {
        switch(iccid_with_fd(fd, &iccid)) {
        case MCTD_RESULT_OK:
            _no_sim_card = false;
            _checked_sim_card = true;
            free(iccid);
            return true;

        case MCTD_RESULT_ERR_NO_DATA:
            sleep(1);
            break;

        default:
            return false;
        }
    }

    _no_sim_card = true;
    _checked_sim_card = true;
    return false;
}

static bool has_sim_card()
{
   int fd;
   bool b = false;

   if (is_calling())
       return true;

   fd = mctd_serial_open();
   if (fd < 0) {
       log_warn("open serial failed.");
       return false;
   }

   b = has_sim_card_with_fd(fd);

   mctd_serial_close(fd);

   return b;
}

static int imei_with_fd(int fd, char **imei)
{
    int result;
    char *cgsn = NULL;
    char *p = NULL;

    /* need echo */
    mctd_atcmd_send(fd, "ATE1", NULL, NULL, NULL);

    result = mctd_atcmd_send_and_echo_with_fd(fd, "AT+CGSN", &cgsn, true, 30);
    if (result != MCTD_RESULT_OK)
        return result;

    if (!cgsn)
        return MCTD_RESULT_ERR_PROCESS;

    p = strtok(cgsn, "\r\n");
    if (!p) {
        result = MCTD_RESULT_ERR_NO_DATA;
        goto bail;
    }

    p = strtok(NULL, "\r\n");
    if (!p) {
        result = MCTD_RESULT_ERR_NO_DATA;
        goto bail;
    }

    free(_imei);
    _imei = malloc(strlen(p) + 1);
    if (!_imei) {
        perror("malloc");
        log_err("malloc failed");
        result = MCTD_RESULT_ERR_MEMORY;
        goto bail;
    }
    strcpy(_imei, p);
    _imei[strlen(p)] = '\0';

    if (!imei)
        goto bail;

    *imei = malloc(strlen(p) + 1);
    if (!*imei) {
        result = MCTD_RESULT_ERR_MEMORY;
        goto bail;
    }
    strcpy(*imei, p);
    (*imei)[strlen(p)] = '\0';

bail:
    free(cgsn);
    return result;
}

static bool is_match_cfun_with_fd(int fd, const char fun)
{
    char *response = NULL;
    char *p = NULL;
    char current_fun;

    if (mctd_atcmd_send(fd, "AT+CFUN?", "+CFUN:", &response, NULL) !=
                                                                MCTD_RESULT_OK)
        return false;

    if (!response)
        return false;

    p = response;
    cutCRandLF(p);
    if (strlen(p) < strlen("+CFUN: x")) {
        free(response);
        return false;
    }

    current_fun = p[strlen(p) - 1];
    free(response);

    return (fun == current_fun);
}

static int set_cfun_with_fd(int fd, char *param)
{
    char buf[16];
    int timeout_sec = 20;

    memset(buf, 0, 16);
    sprintf(buf, "AT+CFUN=%s", param);

    return mctd_atcmd_send(fd, buf, NULL, NULL, &timeout_sec);
}

static int set_cfun(char *param)
{
    int fd;
    int result;

    if (is_calling())
        return MCTD_RESULT_ERR_CALLING;

    fd = mctd_serial_open();
    if (fd < 0) {
        log_warn("open serial failed.");
        return mctd_serial_convert_error(fd);
    }

    result = set_cfun_with_fd(fd, param);

    mctd_serial_close(fd);

    return result;
}

static int set_sms_mode_with_fd(int fd, char *param)
{
    char buf[16];
    int timeout_sec = 20;

    memset(buf, 0, 16);
    snprintf(buf, 16, "AT+CMGF=%s", param);

    return mctd_atcmd_send(fd, buf, NULL, NULL, &timeout_sec);
}

static bool is_need_update_apn_with_fd(int fd, char *apn)
{
    char *p;
    char *response = NULL;
    char *current = NULL;
    bool update = true;
    int result;

    /* check current setting */
    result = mctd_atcmd_send(fd, "AT+CGDCONT?", "+CGDCONT: 1", &response, NULL);
    if (result != MCTD_RESULT_OK)
        goto bail;

    if (!response)
        goto bail;

    p = strtok(response, ",");
    if (!p)
        goto bail;

    p = strtok(NULL, ","); /* IP / IPV6 / IPV4V6 */
    current = strtok(NULL, ","); /* apn */
    if (!current)
        goto bail;

    cutCRandLF(current);
    if (strlen(current) < 3)
        goto bail;

    if (strcmp(current, apn) == 0)
        update = false;

bail:
    free(response);
    return update;
}

static int set_apn_with_fd(int fd, char *apn)
{
    int result;
    char *buf;

    buf = malloc(strlen("AT+CGDCONT=1,\"IP\",") + strlen(apn)
                    +2 /* double quotation x 2 */
                );
    if (!buf)
        return MCTD_RESULT_ERR_MEMORY;

    sprintf(buf, "AT+CGDCONT=1,\"IP\",%s", apn);
    result = mctd_atcmd_send(fd, buf, NULL, NULL, NULL);

    free(buf);
    return result;
}

static int set_apn(char *req_param, char **res_param)
{
    int fd;
    int result;

    UNUSED(res_param);

    if (is_calling())
        return MCTD_RESULT_ERR_CALLING;

    fd = mctd_serial_open();
    if (fd < 0)
        return mctd_serial_convert_error(fd);

    if (!is_need_update_apn_with_fd(fd, req_param)) {
        result = MCTD_RESULT_OK;
        goto bail;
    }

    result = set_cfun_with_fd(fd, "0");
    if (result != MCTD_RESULT_OK)
        goto bail;

    sleep(1);

    result = set_apn_with_fd(fd, req_param);

bail:
    sleep(1);
    set_cfun_with_fd(fd, "1");
    mctd_serial_close(fd);
    return result;
}

static bool is_need_update_auth_with_fd(int fd, char *auth_prot,
                                                char *username, char *password)
{
    char *p;
    char *response = NULL;
    char *current_prot = NULL;
    char *current_username = NULL;
    char *current_password = NULL;
    bool update = true;
    int result;

    /* check current setting */
    result = mctd_atcmd_send(fd, "AT+CGAUTH?", "+CGAUTH: 1", &response, NULL);
    if (result != MCTD_RESULT_OK)
        goto bail;

    if (!response)
        goto bail;

    p = strtok(response, ",");
    if (!p)
        goto bail;

    current_prot = strtok(NULL, ",");
    current_username = strtok(NULL, ",");
    current_password = strtok(NULL, ",");

    if (!current_prot)
        goto bail;

    cutCRandLF(current_prot);

    if (current_username)
        cutCRandLF(current_username);

    if (current_password)
        cutCRandLF(current_password);

    if (strcmp(current_prot, "0") == 0 &&
        strcmp(auth_prot, "\"none\"") == 0)
            update = false;
    else if ((strcmp(current_prot, "1") == 0 &&
              strcmp(auth_prot, "\"pap\"") == 0) ||
             (strcmp(current_prot, "2") == 0 &&
              strcmp(auth_prot, "\"chap\"") == 0))
    {
        if (current_username && current_password && username && password) {
            if (strcmp(current_username, username) == 0 &&
                strcmp(current_password, password) == 0)
            update = false;
        }
    }

bail:
    free(response);
    return update;
}

static int set_auth_with_fd(int fd, char * auth_prot,
                                                char *username, char *password)
{
    int result;
    char *buf;
    char auth_prot_char;
    int cmd_len;

    if (strcmp(auth_prot, "\"pap\"") == 0)
        auth_prot_char = '1';
    else if (strcmp(auth_prot, "\"chap\"") == 0)
        auth_prot_char = '2';
    else if (strcmp(auth_prot, "\"none\"") == 0)
        auth_prot_char = '0';
    else
        return MCTD_RESULT_ERR_PARAM;

    if (auth_prot_char == '0')
        return mctd_atcmd_send(fd, "AT+CGAUTH=1,0", NULL, NULL, NULL);

    if (!username || !password)
        return MCTD_RESULT_ERR_PARAM;

    cmd_len = strlen("AT+CGAUTH=1,2,") + strlen(username)
                                       + strlen(password)
                                       + 1 /* comma */;
    buf = malloc(cmd_len + 1 /* NULL byte */ );
    if (!buf)
        return MCTD_RESULT_ERR_MEMORY;

    sprintf(buf, "AT+CGAUTH=1,%c,%s,%s", auth_prot_char, username, password);
    result = mctd_atcmd_send(fd, buf, NULL, NULL, NULL);
    free(buf);

    return result;
}

static int set_auth(char *auth_prot, char *username, char *password)
{
    int fd;
    int result;

    if (is_calling())
        return MCTD_RESULT_ERR_CALLING;

    fd = mctd_serial_open();
    if (fd < 0)
        return mctd_serial_convert_error(fd);

    if(!is_need_update_auth_with_fd(fd, auth_prot, username, password)) {
        result = MCTD_RESULT_OK;
        goto bail;
    }

    result = set_cfun_with_fd(fd, "0");
    if (result != MCTD_RESULT_OK)
        goto bail;

    sleep(1);

    result = set_auth_with_fd(fd, auth_prot, username, password);

bail:
    sleep(1);
    set_cfun_with_fd(fd, "1");
    mctd_serial_close(fd);
    return result;
}

static int set_apn_and_auth(char *apn,
                            char *auth_prot, char *username, char *password)
{
    int fd;
    int result;
    bool update_auth = true;
    bool update_apn = true;

    if (is_calling())
        return MCTD_RESULT_ERR_CALLING;

    fd = mctd_serial_open();
    if (fd < 0)
        return mctd_serial_convert_error(fd);

    update_apn = is_need_update_apn_with_fd(fd, apn);
    update_auth = is_need_update_auth_with_fd(fd, auth_prot,
                                                        username, password);

    if (!update_apn && !update_auth) {
        result = MCTD_RESULT_OK;
        goto bail;
    }

    result = set_cfun_with_fd(fd, "0");
    if (result != MCTD_RESULT_OK)
        goto bail;

    sleep(1);

    if (update_apn) {
        result = set_apn_with_fd(fd, apn);
        if (result != MCTD_RESULT_OK)
            goto bail;
    }

    if (update_auth) {
        result = set_auth_with_fd(fd, auth_prot, username, password);
    }

bail:
    sleep(1);
    set_cfun_with_fd(fd, "1");
    mctd_serial_close(fd);
    return result;
}


static int operator_id_with_fd(int fd, char **operator_id)
{
    char *p;
    char *response = NULL;
    char *current = NULL;
    int result;

    if (!operator_id) {
        log_err("operator_id is null.");
        return MCTD_RESULT_ERR_PARAM;
    }

    *operator_id = NULL;

    result = mctd_atcmd_send(fd, "AT+COPS?", "+COPS:", &response, NULL);
    if (result != MCTD_RESULT_OK)
        goto bail;

    if (!response) {
        result = MCTD_RESULT_ERR_NO_DATA;
        goto bail;
    }

    p = strtok(response, ",");
    if (!p) {
        result = MCTD_RESULT_ERR_NO_DATA;
        goto bail;
    }

    p = strtok(NULL, ","); /* format */
    if (!p) {
        result = MCTD_RESULT_ERR_NO_DATA;
        goto bail;
    }

    current = strtok(NULL, ","); /* opName */
    if (!current) {
        result = MCTD_RESULT_ERR_NO_DATA;
        goto bail;
    }

    cutCRandLF(current);
    *operator_id = malloc(strlen(current) + 1);
    if (!(*operator_id)) {
        result = MCTD_RESULT_ERR_MEMORY;
        log_err("malloc failed.");
        goto bail;
    }

    strcpy(*operator_id, current);
    result = MCTD_RESULT_OK;

bail:
    free(response);
    return result;
}

static int get_operator_id(char **operator_id)
{
    int fd;
    int result;

    if (is_calling())
        return MCTD_RESULT_ERR_CALLING;

    fd = mctd_serial_open();
    if (fd < 0) {
        log_warn("open serial failed.");
        return mctd_serial_convert_error(fd);
    }

    result = operator_id_with_fd(fd, operator_id);

    mctd_serial_close(fd);

    return result;
}

static bool is_need_update_operator_id_with_fd(int fd, char *operator_id)
{
    char *current = NULL;
    bool update = true;
    int result;

    /* check current operator_id */
    result = operator_id_with_fd(fd, &current);
    if (result != MCTD_RESULT_OK)
        goto bail;

    if (strcmp(current, operator_id) == 0)
        update = false;

bail:
    free(current);
    return update;
}

static int set_operator_id_with_fd(int fd, char *operator_id)
{
    int result;
    char *buf = NULL;
    int timeout_sec;

    if (is_calling())
        return MCTD_RESULT_ERR_CALLING;

    if (!is_need_update_operator_id_with_fd(fd, operator_id)) {
        result = MCTD_RESULT_OK;
        goto bail;
    }

    if (check_param(operator_id, "auto")) {
        buf = malloc(strlen("AT+COPS=0") + 1);
        if (!buf) {
            result = MCTD_RESULT_ERR_MEMORY;
            goto bail;
        }
        sprintf(buf, "AT+COPS=0");
    } else {
        buf = malloc(strlen("AT+COPS=1,2,") + strlen(operator_id) + 1);
        if (!buf) {
            result = MCTD_RESULT_ERR_MEMORY;
            goto bail;
        }
        sprintf(buf, "AT+COPS=1,2,%s", operator_id);
    }

    timeout_sec = 30;
    result = mctd_atcmd_send(fd, buf, NULL, NULL, &timeout_sec);

bail:
    free(buf);
    return result;
}

static int set_operator_id(char *operator_id)
{
    int fd;
    int result;

    if (is_calling())
        return MCTD_RESULT_ERR_CALLING;

    fd = mctd_serial_open();
    if (fd < 0) {
        log_warn("open serial failed.");
        return mctd_serial_convert_error(fd);
    }

    result = set_operator_id_with_fd(fd, operator_id);
    mctd_serial_close(fd);
    return result;
}

/*
 * to use AT^SCFG or AT^SGPICFG
 */
static bool is_need_update_xcfg_with_fd(int fd, const char *cmd,
                                        const char *resp_str, const char *want)
{
    char *p;
    char *response = NULL;
    char *current = NULL;
    bool update = true;
    int result;

    result = mctd_atcmd_send(fd, cmd, resp_str, &response, NULL);

    if (result != MCTD_RESULT_OK || !response)
        goto bail;

    p = strtok(response, ",");
    if (!p)
        goto bail;

    current = strtok(NULL, ",");
    if (!current)
        goto bail;

    cutCRandLF(current);

    if (strcmp(want, current) == 0)
        update = false;

bail:
    free(response);
    return update;
}

static int set_scfg_gpio_mode_asc1_rsv_with_fd(int fd)
{
    if (!is_need_update_xcfg_with_fd(fd,
                        "AT^SCFG?", "^SCFG: \"GPIO/mode/ASC1\"", "\"rsv\""))
        return MCTD_RESULT_OK;

    return mctd_atcmd_send(fd, "AT^SCFG=\"GPIO/mode/ASC1\",\"rsv\"",
                                                            NULL, NULL, NULL);
}

static int set_scfg_gpio_mode_ring0_std_with_fd(int fd)
{
    if (!is_need_update_xcfg_with_fd(fd,
                        "AT^SCFG?", "^SCFG: \"GPIO/mode/RING0\"", "\"std\""))
        return MCTD_RESULT_OK;

    return mctd_atcmd_send(fd, "AT^SCFG=\"GPIO/mode/RING0\",\"std\"",
                                                            NULL, NULL, NULL);
}

static int set_scfg_urc_ringline_asc0_with_fd(int fd)
{
    if (!is_need_update_xcfg_with_fd(fd,
                        "AT^SCFG?", "^SCFG: \"URC/Ringline\"", "\"asc0\""))
        return MCTD_RESULT_OK;

    return mctd_atcmd_send(fd, "AT^SCFG=\"URC/Ringline\", \"asc0\"",
                                                            NULL, NULL, NULL);
}

static int set_scfg_urc_ringline_activetime_1s_with_fd(int fd)
{
    if (!is_need_update_xcfg_with_fd(fd, "AT^SCFG?",
                            "^SCFG: \"URC/Ringline/ActiveTime\"", "\"2\""))
        return MCTD_RESULT_OK;

    return mctd_atcmd_send(fd, "AT^SCFG=\"URC/Ringline/ActiveTime\",\"2\"",
                                                            NULL, NULL, NULL);
}

static int set_sgpicfg_logic_ring0_low_with_fd(int fd)
{
    if (!is_need_update_xcfg_with_fd(fd,
                        "AT^SGPICFG?", "^SGPICFG: \"Logic/Ring0\"", "\"low\""))
        return MCTD_RESULT_OK;

    return mctd_atcmd_send(fd, "AT^SGPICFG=\"Logic/Ring0\",\"low\"",
                                                            NULL, NULL, NULL);
}

static int set_sgpicfg_port_wakein_gpio25_with_fd(int fd)
{
    if (!is_need_update_xcfg_with_fd(fd, "AT^SGPICFG?",
                                    "^SGPICFG: \"port/WakeIn\"", "\"GPIO25\""))
        return MCTD_RESULT_OK;

    return mctd_atcmd_send(fd, "AT^SGPICFG=\"port/WakeIn\",\"GPIO25\"",
                                                            NULL, NULL, NULL);
}

static int set_scfg_gprs_autoattach_with_fd(int fd, bool enable_auto_attach)
{
    if (!is_need_update_xcfg_with_fd(fd,
                    "AT^SCFG?", "^SCFG: \"GPRS/AutoAttach\"",
                    enable_auto_attach?"\"enable\"":"\"disable\""))
        return MCTD_RESULT_OK;

    if (enable_auto_attach)
        return mctd_atcmd_send(fd, "AT^SCFG=\"GPRS/AutoAttach\",\"enable\"",
                                                            NULL, NULL, NULL);
    else
        return mctd_atcmd_send(fd, "AT^SCFG=\"GPRS/AutoAttach\",\"disable\"",
                                                            NULL, NULL, NULL);
}

static int set_auto_attach(bool enable_auto_attach)
{
    int fd;
    int result;

    if (is_calling())
        return MCTD_RESULT_ERR_CALLING;

    fd = mctd_serial_open();
    if (fd < 0) {
        log_warn("open serial failed.");
        return mctd_serial_convert_error(fd);
    }
    result = set_scfg_gprs_autoattach_with_fd(fd, enable_auto_attach);
    mctd_serial_close(fd);
    return result;
}

static bool is_need_update_fix_profile_with_fd(int fd, char *fix_profile)
{
    char *p;
    char *response = NULL;
    char *current = NULL;
    bool update = true;
    bool is_manual = false;
    int result;

    /* check AutoSelect setting */
    result = mctd_atcmd_send(fd, "AT^SCFG?", "^SCFG: \"MEopMode/Prov/AutoSelect\"",
                                                            &response, NULL);
    if (result != MCTD_RESULT_OK)
        goto bail;

    if (!response)
        goto bail;

    p = strtok(response, ",");
    if (!p)
        goto bail;

    current = strtok(NULL, ",");
    if (!current)
        goto bail;

    cutCRandLF(current);
    is_manual = (strcmp(current, "\"off\"") == 0);

    if (!fix_profile) {
        update = is_manual;
        goto bail;
    } else if (!is_manual)
        goto bail;

    free(response);
    response = NULL;

    /* check MEopMode/Prov/Cfg */
    result = mctd_atcmd_send(fd, "AT^SCFG?", "^SCFG: \"MEopMode/Prov/Cfg\"",
                                                            &response, NULL);

    if (result != MCTD_RESULT_OK || !response)
        goto bail;

    p = strtok(response, ",");
    if (!p)
        goto bail;

    current = strtok(NULL, ",");
    if (!current)
        goto bail;

    cutCRandLF(current);

    if (strcmp(current, fix_profile) == 0)
        update = false;

bail:
    free(response);
    return update;
}

static void wait_fixed_profile(int fd)
{
    int timeout = 1;

    /* if AT command failed, modem rebooting, and need wait ~SYSSTART */
    if (mctd_atcmd_send(fd, "AT", NULL, NULL, &timeout) != MCTD_RESULT_OK) {
        log_info("wait reboot.");
        mctd_atcmd_wait_urc_with_fd_and_timeout(fd, URC_SYSSTART, 60);
    }
}

static int set_fix_profile(char *fix_profile)
{
    int fd;
    int result;
    char *buf = NULL;

    if (is_calling())
        return MCTD_RESULT_ERR_CALLING;

    fd = mctd_serial_open();
    if (fd < 0)
        return mctd_serial_convert_error(fd);

    if (!is_need_update_fix_profile_with_fd(fd, fix_profile)) {
        result = MCTD_RESULT_OK;
        goto bail;
    }

    if (!fix_profile) {
        mctd_atcmd_send(fd, "AT^SCFG=\"MEopMode/Prov/AutoSelect\",\"on\"",
                                                            NULL, NULL, NULL);
        result = MCTD_RESULT_OK;
    } else {
        buf = malloc(strlen("AT^SCFG=\"MEopMode/Prov/Cfg\",")
                    + strlen(fix_profile)
                    + 2 /* double quotation x 2 */ );
        if (!buf) {
            result = MCTD_RESULT_ERR_MEMORY;
            goto bail;
        }
        sprintf(buf, "AT^SCFG=\"MEopMode/Prov/Cfg\",%s", fix_profile);

        result = mctd_atcmd_send(fd, buf, NULL, NULL, NULL);
        if (result != MCTD_RESULT_OK) {
            free(buf);
            goto bail;
        }

        wait_fixed_profile(fd);

        result = mctd_atcmd_send(fd,
                                "AT^SCFG=\"MEopMode/Prov/AutoSelect\",\"off\"",
                                                            NULL, NULL, NULL);
        free(buf);
    }

    wait_fixed_profile(fd);

bail:
    mctd_serial_close(fd);

    return result;
}

static int dial(char *req_param, char **res_param)
{
    UNUSED(req_param);
    UNUSED(res_param);

    if (is_wvdial_running())
        return MCTD_RESULT_OK;
    if (_multiplex == MUX_STATUS_DISABLE) {
        strcpy(at_port, SERIAL_TTY);
    } else {
        strcpy(at_port, MUX_AT_TTY);
    }
    bin_execution(true, "wvdial", NULL);
    return MCTD_RESULT_OK;
}

static int hangup(char *req_param, char **res_param)
{
    int i;

    UNUSED(req_param);
    UNUSED(res_param);

    if (!is_wvdial_running())
        return MCTD_RESULT_OK;

    bin_execution(false, "poff.wvdial", NULL);

    for(i = 0; i < 10; i++) {
        if (!is_wvdial_running()) {
            strcpy(at_port, SERIAL_TTY);
            return MCTD_RESULT_OK;
            break;
        }
        sleep(3);
    }

    return MCTD_RESULT_ERR_CALLING;
}

/**
 * @note AT+CNUM
 *       Response(s)
 *       [+CNUM: [<alpha>], <number>, <type>]
 *       [+CNUM: ... ]
 *       use <number> at 1st element.
 *
 */
static int get_phone_number(char *req_param, char **res_param)
{
    char *temp = NULL;
    char *p;
    int result;

    UNUSED(req_param);

    if (is_calling())
        return MCTD_RESULT_ERR_CALLING;

    result = mctd_atcmd_send_oneshot("AT+CNUM", "+CNUM", &temp);
    if (result != MCTD_RESULT_OK)
        goto bail;

    if (!temp) {
        result = MCTD_RESULT_ERR_NO_DATA;
        goto bail;
    }

    /* phase phone number */
    p = strtok(temp, ",");
    p = strtok(NULL, ",");

    if (res_param) {
        *res_param = malloc(strlen(p)-1);
        if (!*res_param) {
            result = MCTD_RESULT_ERR_MEMORY;
            goto bail;
        }
        /* without double quotations */
        memset(*res_param, 0, strlen(p)-1);
        strncpy(*res_param, &p[1], strlen(p) - 2);
    }

    free(_phone_number);
    _phone_number = malloc(strlen(p)-1);
    if (!_phone_number) {
        result = MCTD_RESULT_ERR_MEMORY;
        goto bail;
    }
    /* without double quotations */
    memset(_phone_number, 0, strlen(p)-1);
    strncpy(_phone_number, &p[1], strlen(p) - 2);

bail:
    free(temp);
    return result;
}

/*
 * @note AT+CESQ
 *       +CESQ: <rxlev>, <ber>, <rscp>, <ecno>, <rsrq>, <rsrp>
 *       use <rsrq>
 *         0 rsrq < -19.5 dB
 *         1 -19.5 dB <= rsrq < -19   dB
 *         2 -19   dB <= rsrq < -18.5 dB
 *        32 - 4   dB <= rsrq < - 3.5 dB
 *        33 - 3.5 dB <= rsrq < - 3   dB
 *        34 - 3   dB <= rsrq
 *       255 not known or not detectable
 */
static int get_signal_quality(char *req_param, char **res_param)
{
    char *temp = NULL;
    char *p;
    int result;
    int i;

    UNUSED(req_param);

    if (is_calling())
        return MCTD_RESULT_ERR_CALLING;

    result = mctd_atcmd_send_oneshot("AT+CESQ", "+CESQ", &temp);
    if (result != MCTD_RESULT_OK)
        goto bail;

    if (!temp) {
        result = MCTD_RESULT_ERR_NO_DATA;
        goto bail;
    }

    /* phase signal quality */
    p = strtok(temp, ",");
    if (!p) {
        result = MCTD_RESULT_ERR_PROCESS;
        goto bail;
    }

    /* go 5th element <rsrq> */
    for (i = 0; i < 4; i++) {
        p = strtok(NULL, ",");
        if (!p) {
            result = MCTD_RESULT_ERR_PROCESS;
            goto bail;
        }
    }

    if (res_param) {
        *res_param = malloc(strlen(p) + 1);
        if (!*res_param) {
            perror("malloc");
            result = MCTD_RESULT_ERR_MEMORY;
            goto bail;
        }
        memset(*res_param, 0, strlen(p) + 1);
        strncpy(*res_param, p, strlen(p));
    }
    _rsrq = atoi(p);

    p = strtok(NULL, ",");
    if (p) {
        cutCRandLF(p);
        _rsrp = atoi(p);
    }

bail:
    free(temp);
    return result;
}

static int set_lte_gpio25(bool high)
{
    int fd;

    fd = open(EMS31_GPIO25_PATH, O_WRONLY);
    if (fd == -1)
        return MCTD_RESULT_ERR_PROCESS;

    write(fd, high?"1":"0", 1);

    close(fd);

    return MCTD_RESULT_OK;
}

static int activate(char *req_param, char **res_param)
{
    UNUSED(req_param);
    UNUSED(res_param);
    _is_suspend = false;

    /* set GPIO25 high */
    return set_lte_gpio25(true);
}

static int deactivate(char *req_param, char **res_param)
{
    UNUSED(req_param);
    UNUSED(res_param);
    _is_suspend = true;

    mctd_atcmd_flush();

    /* set GPIO25 low */
    return set_lte_gpio25(false);
}

static int ems_31_power_ctl(bool on)
{
    int fd;

    fd = open(EMS31_PWR_CTL_PATH, O_WRONLY);
    if (fd == -1) {
        perror("open ems31 pwr ctl");
        return MCTD_RESULT_ERR_PROCESS;
    }

    write(fd, on?"1":"0", 1);

    close(fd);

    return MCTD_RESULT_OK;
}

/* prototype for poweron() */
static void startup_modem();
static void startup(const char *conf_file_path);

static int poweron(char *req_param, char **res_param)
{
    int result;
    int i;

    UNUSED(req_param);
    UNUSED(res_param);

    if (is_wvdial_running())
        return MCTD_RESULT_ERR_CALLING;

    /* power supply on for EMS31 and ignition */
    result = ems_31_power_ctl(true);
    if (result != MCTD_RESULT_OK) {
        log_warn("poweron modem failed.");
        return result;
    }

    activate(NULL, NULL);

    for (i = 0; i < 20; i++) {
        if (set_serial_baudrate() == MCTD_RESULT_OK)
            break;
    }
    strcpy(ppp_port, SERIAL_TTY);
    strcpy(at_port, SERIAL_TTY);
    startup_modem();
    startup(STARTUP_CONFIG_PATH);

    return result;
}

static int power_supply_off(char *req_param, char **res_param)
{
    int result;

    UNUSED(req_param);
    UNUSED(res_param);

    result = ems_31_power_ctl(false);
    if (result != MCTD_RESULT_OK)
        log_warn("poweroff modem failed.");

    return result;
}

static int poweroff(char *req_param, char **res_param)
{
    hangup(req_param, res_param);

    if (mctd_atcmd_send_oneshot("AT", NULL, NULL) == MCTD_RESULT_OK)
        mctd_atcmd_send_and_wait_oneshot("AT^SMSO", URC_SHUTDOWN, 60);

    /* power supply off for EMS31 */
    return power_supply_off(req_param, res_param);
}

static int wwan_force_restart(char *req_param, char **res_param)
{
    int result;
    bool was_calling;

    log_info("start wwan-force-restart");

    was_calling = is_wvdial_running();

    poweroff(req_param, res_param);

    /* wait some seconds */
    sleep(5);

    result = poweron(req_param, res_param);
    if (result != MCTD_RESULT_OK) {
        log_info("wwan-force-restart failed");
        return result;
    }

    if (was_calling)
        dial(NULL, NULL);

    log_info("finish wwan-force-restart");
    return MCTD_RESULT_OK;
}

#define PSM_MODE_DISABLE "0"
#define PSM_MODE_ENABLE  "1"
#define PSM_MODE_DEFAULT "2"

#define PSM_TAU_10MIN   "000"
#define PSM_TAU_1HOUR   "001"
#define PSM_TAU_10HOURS "010"
#define PSM_TAU_2SEC    "011"
#define PSM_TAU_30SEC   "100"
#define PSM_TAU_1MIN    "101"

#define PSM_ACT_2SEC      "000"
#define PSM_ACT_1MIN      "001"
#define PSM_ACT_DECIHOURS "010"

typedef enum {
    PSM_UNIT_SEC,
    PSM_UNIT_MIN,
    PSM_UNIT_HOUR,
    PSM_UNIT_UNKNOWN,
} PSM_UNIT_T;

static char bit_char(long num, long pos)
{
    return num & pos ? '1' : '0';
}

static void print_bin_psm(char *at_str, const char *unit, long num)
{
    memset(at_str, 0, 9);
    sprintf(at_str, "\"%s%c%c%c%c%c\"", unit, bit_char(num, 16), bit_char(num, 8),
                        bit_char(num, 4), bit_char(num, 2), bit_char(num, 1));
}

static int get_unit(const char *cmd_str, const int len)
{
    if (cmd_str[len-1] == 's')
        return PSM_UNIT_SEC;

    if (cmd_str[len-1] == 'm')
        return PSM_UNIT_MIN;

    if (cmd_str[len-1] == 'h')
        return PSM_UNIT_HOUR;

    return PSM_UNIT_UNKNOWN;
}

static int psm_val_and_unit(const char *cmd_str, int *unit)
{
    int len;
    char *end;
    errno = 0;
    long num;

    if (!cmd_str)
        return -1;

    len = strlen(cmd_str);
    if (len < 2 || len > 5)
        return -1;

    num = strtol(cmd_str, &end, 10);
    if ((num == LONG_MIN || num == LONG_MAX) && errno) {
        return -1;
    }

    if (end - cmd_str != len - 1) {
        return -1;
    }

    *unit = get_unit(end ,1);
    return num;
}

static int psm_tau_to_at(const char *cmd_str, char *at_str)
{
    long num;
    int unit;

    num = psm_val_and_unit(cmd_str, &unit);
    if (num == -1)
        return MCTD_RESULT_ERR_PARAM;

    switch(unit) {
    case PSM_UNIT_SEC:
        if (num <= MCTD_PSM_MAX_BIT*MCTD_PSM_2SEC) {
            num = num / 2;
            print_bin_psm(at_str, PSM_TAU_2SEC, num);
        } else {
            num = num / 30;
            print_bin_psm(at_str, PSM_TAU_30SEC, num);
        }
        break;
    case PSM_UNIT_MIN:
        if (num <= MCTD_PSM_MAX_BIT)
            print_bin_psm(at_str, PSM_TAU_1MIN, num);
        else {
            num = num / 10;
            print_bin_psm(at_str, PSM_TAU_10MIN, num);
        }
        break;
    case PSM_UNIT_HOUR:
        if (num <= MCTD_PSM_MAX_BIT)
            print_bin_psm(at_str, PSM_TAU_1HOUR, num);
        else {
            num = num / 10;
            print_bin_psm(at_str, PSM_TAU_10HOURS, num);
        }
        break;
    default:
        return MCTD_RESULT_ERR_PARAM;
    }

    return MCTD_RESULT_OK;
}

static int psm_act_to_at(const char *cmd_str, char *at_str)
{
    long num;
    int unit;

    num = psm_val_and_unit(cmd_str, &unit);
    if (num == -1)
        return MCTD_RESULT_ERR_PARAM;

    switch(unit) {
    case PSM_UNIT_SEC:
        num = num / 2;
        print_bin_psm(at_str, PSM_ACT_2SEC, num);
        break;
    case PSM_UNIT_MIN:
        if (num <= MCTD_PSM_MAX_BIT)
            print_bin_psm(at_str, PSM_ACT_1MIN, num);
        else {
            num = num / 6;
            print_bin_psm(at_str, PSM_ACT_DECIHOURS, num);
        }
        break;
    default:
        return MCTD_RESULT_ERR_PARAM;
    }

    return MCTD_RESULT_OK;
}

static int set_psm(char *req_param, char **res_param)
{
    int fd;
    int result;

    char *mode_req     = NULL;
    char *tau_req      = NULL;
    char *act_req      = NULL;

    char tau_req_str[11];
    char act_req_str[11];

    char *p            = NULL;
    char *response     = NULL;
    char *mode_current = NULL;
    char *tau_current  = NULL;
    char *act_current  = NULL;
    bool update        = false;
    char cmd_str[35];

    UNUSED(res_param);

    if (is_calling())
        return MCTD_RESULT_ERR_CALLING;

    fd = mctd_serial_open();
    if (fd < 0)
        return mctd_serial_convert_error(fd);

    if (req_param)
        mode_req = strtok(req_param, "\t");
    else {
        result = MCTD_RESULT_ERR_PARAM;
        goto bail;
    }

    /* always do setting default */
    if ((mode_req && (strcmp(mode_req, MCTD_STR_DEFAULT) == 0)) ||
        (strcmp(req_param, MCTD_STR_DEFAULT) == 0)) {
        result = mctd_atcmd_send(fd, "AT+CPSMS=2", NULL, NULL, NULL);
        goto bail;
    }

    /* parse request parameter */
    if (mode_req) {
        if (strcmp(mode_req, MCTD_STR_ENABLE) == 0) {
            tau_req = strtok(NULL, "\t");
            act_req = strtok(NULL, "\t");
            if (!tau_req || !act_req) {
                result = MCTD_RESULT_ERR_PARAM;
                goto bail;
            }
            result = psm_tau_to_at(tau_req, tau_req_str);
            if (result != MCTD_RESULT_OK)
                goto bail;

            result = psm_act_to_at(act_req, act_req_str);
            if (result != MCTD_RESULT_OK)
                goto bail;
        } else if (strcmp(mode_req, MCTD_STR_DISABLE) == 0) {
            mode_req = req_param;
        }
    } else {
       result = MCTD_RESULT_ERR_PARAM;
       goto bail;
    }

    /* parse current setting */
    result = mctd_atcmd_send(fd, "AT+CPSMS?", "+CPSMS", &response, NULL);
    if (result != MCTD_RESULT_OK)
        goto bail;

    if (!response)
        update = true;
    else {
        p = strtok(response, ",");
        if (p) {
            mode_current = p;
            tau_current = strtok(NULL, ",");

            act_current = strtok(NULL, ",");
            cutCRandLF(act_current);

            p = strtok(mode_current, " ");
            mode_current = strtok(NULL, " ");
        }
    }

    /* compare current and request settings */
    if (update) {
        /* always need update */
    } else if (!mode_current)
        update = true;
    else {
        if ((strcmp(mode_req, MCTD_STR_ENABLE) == 0) &&
            (strcmp(mode_current, "1") == 0)) {
            /* enable - enable */
            if (!tau_current || !act_current)
                update = true;
            else {
                /* compare tau and act-time */
                if ((strcmp(tau_req_str, tau_current) != 0) ||
                    (strcmp(act_req_str, act_current) != 0))
                    update = true;
            }
        } else if ((strcmp(mode_req, MCTD_STR_DISABLE) == 0) &&
                   (strcmp(mode_current, "0") == 0))
            /* disable - disable */
            update = false;
        else
            update = true;
    }

    if (!update) {
        result = MCTD_RESULT_OK;
        goto bail;
    }

    if (strcmp(mode_req, MCTD_STR_ENABLE) == 0) {
        sprintf(cmd_str, "AT+CPSMS=1,,,%s,%s", tau_req_str, act_req_str);
        result = mctd_atcmd_send(fd, cmd_str, NULL, NULL, NULL);
    } else
        result = mctd_atcmd_send(fd, "AT+CPSMS=0", NULL, NULL, NULL);

bail:
    free(response);
    mctd_serial_close(fd);
    return result;
}

const char* table_pcl[][2] = {
    { "\"0000\"",    "5.12" },
    { "\"0001\"",   "10.24" },
    { "\"0010\"",   "20.48" },
    { "\"0011\"",   "40.96" },
    { "\"0100\"",   "61.44" },
    { "\"0101\"",   "81.92" },
    { "\"0110\"",  "102.40" },
    { "\"0111\"",  "122.88" },
    { "\"1000\"",  "143.36" },
    { "\"1001\"",  "163.84" },
    { "\"1010\"",  "327.68" },
    { "\"1011\"",  "655.36" },
    { "\"1100\"", "1310.72" },
    { "\"1101\"", "2621.44" },
    { NULL,       NULL      },
};

const char* table_ptw[][2] = {
    { "\"0000\"",  "1.28" },
    { "\"0001\"",  "2.56" },
    { "\"0010\"",  "3.84" },
    { "\"0011\"",  "5.12" },
    { "\"0100\"",  "6.40" },
    { "\"0101\"",  "7.68" },
    { "\"0110\"",  "8.96" },
    { "\"0111\"", "10.24" },
    { "\"1000\"", "11.52" },
    { "\"1001\"", "12.80" },
    { "\"1010\"", "14.08" },
    { "\"1011\"", "15.36" },
    { "\"1100\"", "16.64" },
    { "\"1101\"", "17.92" },
    { "\"1110\"", "19.20" },
    { "\"1111\"", "20.48" },
    { NULL,       NULL    },
};

static int search_edrx_pcl(const char *str_cmd, int *pos)
{
    if (!str_cmd)
        return MCTD_RESULT_ERR_PARAM;

    for (*pos=0; ;(*pos)++) {
        if (!table_pcl[*pos][0])
            return MCTD_RESULT_ERR_PARAM;

        if (strcmp(str_cmd, table_pcl[*pos][1]) == 0)
            return MCTD_RESULT_OK;
    }
    return MCTD_RESULT_ERR_PARAM;
}

static int search_edrx_ptw(const char *str_cmd, int *pos)
{
    if (!str_cmd)
        return MCTD_RESULT_ERR_PARAM;

    for (*pos=0; ;(*pos)++) {
        if (!table_ptw[*pos][0])
            return MCTD_RESULT_ERR_PARAM;

        if (strcmp(str_cmd, table_ptw[*pos][1]) == 0)
            return MCTD_RESULT_OK;
    }
    return MCTD_RESULT_ERR_PARAM;
}

static int set_edrx(char *req_param, char **res_param)
{
    UNUSED(res_param);

    int fd;
    int result;
    int pcl_pos = -1;
    int ptw_pos = -1;
    char *mode_req     = NULL;
    char *pcl_req      = NULL;
    char *ptw_req      = NULL;

    char *p            = NULL;
    char *response     = NULL;
    char *mode_current = NULL;
    char *pcl_current  = NULL;
    char *ptw_current  = NULL;
    bool update        = false;
    char cmd_str[28];

    if (is_calling())
        return MCTD_RESULT_ERR_CALLING;

    fd = mctd_serial_open();
    if (fd < 0)
        return mctd_serial_convert_error(fd);

    if (req_param)
        mode_req = strtok(req_param, "\t");
    else {
        result = MCTD_RESULT_ERR_PARAM;
        goto bail;
    }

    /* always do setting default */
    if ((mode_req && (strcmp(mode_req, MCTD_STR_DEFAULT) == 0)) ||
        (strcmp(req_param, MCTD_STR_DEFAULT) == 0)) {
        result = mctd_atcmd_send(fd, "AT^SEDRXS=3", NULL, NULL, NULL);
        goto bail;
    }

    /* parse request parameter */
    if (mode_req) {
        if (strcmp(mode_req, MCTD_STR_ENABLE) == 0) {
            pcl_req = strtok(NULL, "\t");
            ptw_req = strtok(NULL, "\t");
            if (!pcl_req || !ptw_req) {
                result = MCTD_RESULT_ERR_PARAM;
                goto bail;
            }
        } else if (strcmp(mode_req, MCTD_STR_DISABLE) == 0) {
            mode_req = req_param;
        }
    } else {
        result = MCTD_RESULT_ERR_PARAM;
        goto bail;
    }

    /* parse current setting */
    result = mctd_atcmd_send(fd, "AT^SEDRXS?", "^SEDRXS", &response, NULL);
    if (result != MCTD_RESULT_OK)
        goto bail;

    if (!response)
        update = true;
    else {
        p = strtok(response, ",");
        if (p) {
            pcl_current = strtok(NULL, ",");
            ptw_current = strtok(NULL, ",");
            mode_current = strtok(NULL, ",");
            cutCRandLF(mode_current);
        }
    }

    /* compare current and request settings */
    if (update) {
        /* always need update */
    } else if (!mode_current)
        update = true;
    else {
        if ((strcmp(mode_req, MCTD_STR_ENABLE) == 0) &&
            ((strcmp(mode_current, "1") == 0) ||
             (strcmp(mode_current, "2") == 0))) {
            /* enable - enable */
            if (!pcl_current || !ptw_current)
                update = true;
            else {
                /* compare pcl and ptw */
                update = true;
                result = search_edrx_pcl(pcl_req, &pcl_pos);
                if (result != MCTD_RESULT_OK)
                    goto bail;

                if (strcmp(pcl_current, table_pcl[pcl_pos][0]) == 0)
                     update = false;

                if (!update) {
                    update = true;
                    result = search_edrx_ptw(ptw_req, &ptw_pos);
                    if (result != MCTD_RESULT_OK)
                        goto bail;

                    if (strcmp(ptw_current, table_ptw[ptw_pos][0]) == 0)
                        update = false;
                }
            }
        } else if ((strcmp(mode_req, MCTD_STR_DISABLE) == 0) &&
                   (strcmp(mode_current, "0") == 0))
            /* disable - disable */
            update = false;
        else
            update = true;
    }

    if (!update) {
        result = MCTD_RESULT_OK;
        goto bail;
    }

    if (strcmp(mode_req, MCTD_STR_ENABLE) == 0) {
        if (pcl_pos == -1) {
            result = search_edrx_pcl(pcl_req, &pcl_pos);
            if (result != MCTD_RESULT_OK)
                goto bail;
	}
	if (ptw_pos == -1) {
            result = search_edrx_ptw(ptw_req, &ptw_pos);
            if (result != MCTD_RESULT_OK)
                goto bail;
        }
        sprintf(cmd_str, "AT^SEDRXS=2,4,%s,%s", table_pcl[pcl_pos][0],
                                                table_ptw[ptw_pos][0]);
        result = mctd_atcmd_send(fd, cmd_str, NULL, NULL, NULL);
    } else
        result = mctd_atcmd_send(fd, "AT^SEDRXS=0", NULL, NULL, NULL);

bail:
    free(response);
    mctd_serial_close(fd);
    return result;
}

static int set_suspend(char *req_param, char **res_param)
{
    int fd;
    int result;
    char *response = NULL;
    char *current_mode = NULL;
    char *p;

    UNUSED(res_param);

    if (is_calling())
        return MCTD_RESULT_ERR_CALLING;

    fd = mctd_serial_open();
    if (fd < 0)
        return mctd_serial_convert_error(fd);

    /* read current setting */
    result = mctd_atcmd_send(fd, "AT^SCFG?", "^SCFG: \"MEopMode/PowerMgmt/Suspend\"",
                                                            &response, NULL);
    if (result != MCTD_RESULT_OK)
        goto bail;

    if (response) {
        p = strtok(response, ",");
        if (p)
            current_mode = strtok(NULL, ",");
    }

    if (strcmp(req_param, MCTD_STR_ENABLE) == 0) {
        if (current_mode && strcmp(current_mode, "\"1\"") == 0)
            goto bail;

        result = mctd_atcmd_send(fd,
                        "AT^SCFG=\"MEopMode/PowerMgmt/Suspend\",\"1\",\"1\"",
                                                            NULL, NULL, NULL);
    } else if (strcmp(req_param, MCTD_STR_DISABLE) == 0) {
        if (current_mode && strcmp(current_mode, "\"0\"") == 0)
            goto bail;

        result = mctd_atcmd_send(fd,
                        "AT^SCFG=\"MEopMode/PowerMgmt/Suspend\",\"0\",\"1\"",
                                                            NULL, NULL, NULL);
    }

bail:
    free(response);
    mctd_serial_close(fd);
    return result;
}

static int set_sleep(char *req_param, char **res_param)
{
    UNUSED(res_param);

    int fd;
    int result;
    char *response     = NULL;
    char *mode_current = NULL;
    char *msec_current = NULL;
    char *p;
    char *mode_req     = NULL;
    char *msec_req     = NULL;
    char cmd_str[15];

    UNUSED(res_param);

    if (is_calling())
        return MCTD_RESULT_ERR_CALLING;

    fd = mctd_serial_open();
    if (fd < 0)
        return mctd_serial_convert_error(fd);

    /* parse request parameter */
    if (req_param)
        mode_req = strtok(req_param, "\t");
    else {
        result = MCTD_RESULT_ERR_PARAM;
        goto bail;
    }
    if (mode_req) {
        if (strcmp(mode_req, MCTD_STR_ENABLE) == 0) {
            msec_req = strtok(NULL, "\t");
            if (!msec_req) {
                result = MCTD_RESULT_ERR_PARAM;
                goto bail;
            }
        }
    } else
       mode_req = req_param;

    /* parse current setting */
    result = mctd_atcmd_send(fd, "AT^SPOW?", "^SPOW", &response, NULL);
    if (result != MCTD_RESULT_OK)
        goto bail;

    if (response) {
        p = strtok(response, ",");
        if (p) {
            mode_current = p;
            msec_current = strtok(NULL, ",");
            cutCRandLF(msec_current);

            p = strtok(mode_current, " ");
            mode_current = strtok(NULL, " ");
        }
    }

    if (mode_current) {
        if ((strcmp(mode_current, "1") == 0) &&
            (strcmp(mode_req, MCTD_STR_DISABLE) == 0)) {
            /* disable - disable */
            result = MCTD_RESULT_OK;
            goto bail;
        } else if ((strcmp(mode_current, "2") == 0) &&
                   msec_req /* mode_req == ENABLE */) {
            /* enable - enable */
            if (msec_current && strcmp(msec_req, msec_current) == 0) {
                result = MCTD_RESULT_OK;
                goto bail;
            }
        }
    }

    if (strcmp(mode_req, MCTD_STR_DISABLE) == 0) {
        result = mctd_atcmd_send(fd, "AT^SPOW=1,0", NULL, NULL, NULL);
    } else if (strcmp(mode_req, MCTD_STR_ENABLE) == 0) {
        memset(cmd_str, 0, 15);
        sprintf(cmd_str, "AT^SPOW=2,%s", msec_req);
        result = mctd_atcmd_send(fd, cmd_str, NULL, NULL, NULL);
    } else {
        result = MCTD_RESULT_ERR_PARAM;
        goto bail;
    }

bail:
    free(response);
    mctd_serial_close(fd);
    return result;
}

static int send_at_command(char *req_param, char **res_param)
{
    UNUSED(res_param);

    if (is_calling())
        return MCTD_RESULT_ERR_CALLING;

    return mctd_atcmd_send_oneshot(req_param, NULL, NULL);
}

static int send_at_command_echo(char *req_param, char **res_param)
{
    if (is_calling())
        return MCTD_RESULT_ERR_CALLING;

    return mctd_atcmd_send_and_echo_oneshot(req_param, res_param);
}

static int sim_slot_with_fd(int fd, int *slot)
{
    int result;
    char *response = NULL;
    char *current_slot = NULL;
    char *p;

    /* read current setting */
    result = mctd_atcmd_send(fd, "AT^SCFG?", "^SCFG: \"SIM/CS\"", &response, NULL);
    if (result != MCTD_RESULT_OK)
        return result;

    if (!response)
        return MCTD_RESULT_ERR_PARAM;

    p = strtok(response, ",");
    if (p) {
        current_slot = strtok(NULL, ",");
        cutCRandLF(current_slot);
    }

    if (!current_slot)
        result = MCTD_RESULT_ERR_NO_DATA;
    else if (strcmp(current_slot, "\"SIM1\"") == 0) {
        *slot = 1;
        _sim_slot = 1;
    } else if (strcmp(current_slot, "\"SIM2\"") == 0) {
        *slot = 2;
        _sim_slot = 2;
    } else
        result = MCTD_RESULT_ERR_PARAM;

    free(response);
    return result;
}

static int set_sim_slot(char *req_param, char **res_param)
{
    int fd;
    int result;
    int slot;

    UNUSED(res_param);

    if (is_calling())
        return MCTD_RESULT_ERR_CALLING;

    fd = mctd_serial_open();
    if (fd < 0)
        return mctd_serial_convert_error(fd);

    /* read current setting */
    result = sim_slot_with_fd(fd, &slot);
    if (result != MCTD_RESULT_OK)
        goto bail;

    if (strcmp(req_param, "1") == 0) {
        if (slot == 1)
            goto bail;

        result = mctd_atcmd_send(fd, "AT^SCFG=\"SIM/CS\",\"SIM1\"", NULL, NULL, NULL);
        if (result == MCTD_RESULT_OK)
            _sim_slot = 1;
    } else if (strcmp(req_param, "2") == 0) {
        if (slot == 2)
            goto bail;

        result = mctd_atcmd_send(fd, "AT^SCFG=\"SIM/CS\",\"SIM2\"", NULL, NULL, NULL);
        if (result == MCTD_RESULT_OK)
           _sim_slot = 2;
    } else
        result = MCTD_RESULT_ERR_PARAM;

bail:
    mctd_serial_close(fd);
    return result;
}

static int get_datetime_with_fd(int fd, char **res_param)
{
    int result;
    char *response = NULL;

    result = mctd_atcmd_send(fd, "AT+CCLK?", "+CCLK:", &response, NULL);
    if (result != MCTD_RESULT_OK)
        goto bail;

    if (!response) {
        result = MCTD_RESULT_ERR_PROCESS;
        goto bail;
    }

    cutCRandLF(response);
    /* cut double quotation */
    response[strlen(response) - 1] = '\0';

    *res_param = malloc(strlen(&(response[8])) + 1);
    if (!*res_param) {
        result = MCTD_RESULT_ERR_MEMORY;
        goto bail;
    }

    sprintf(*res_param, "%s", &(response[8]));

bail:
    free(response);
    return result;
}

static int update_datetime(char *req_param, char **res_param)
{
    char req[256];
    int fd;
    int result;

    if (is_calling())
        return MCTD_RESULT_ERR_CALLING;

    fd = mctd_serial_open();
    if (fd < 0)
        return mctd_serial_convert_error(fd);

    memset(req, 0, 256);
    snprintf(req, 256, "AT^SISX=\"Ntp\",1,%s", req_param);

    result = mctd_atcmd_send(fd, req, NULL, NULL, NULL);
    if (result != MCTD_RESULT_OK)
        goto bail;

    result = get_datetime_with_fd(fd, res_param);

bail:
    mctd_serial_close(fd);
    return result;
}

static int get_datetime(char *req_param, char **res_param)
{
    int result;
    int fd;

    UNUSED(req_param);

    if (is_calling())
        return MCTD_RESULT_ERR_CALLING;

    fd = mctd_serial_open();
    if (fd < 0)
        return mctd_serial_convert_error(fd);

    result = get_datetime_with_fd(fd, res_param);

    mctd_serial_close(fd);

    return result;
}

static int list_sms_with_fd(int fd, const char *option, char **res_param)
{
    int result, response_len;
    char *response = NULL;
    char *command = "AT+CMGL";
    char buf[MCTD_AT_BUF_SIZE] = {0};

    snprintf(buf, MCTD_AT_BUF_SIZE, "%s=%s", command, option ? option : "ALL");

    result = mctd_atcmd_send_and_echo_with_fd(fd, buf, &response, false, 1);

    if (result != MCTD_RESULT_OK)
        goto bail;

    if (!response) {
        result = MCTD_RESULT_ERR_PROCESS;
        goto bail;
    }

    /* cut response "OK" and line breaks before and after */
    response_len = strlen(response);
    if (response_len >= AT_RESPONSE_OK_LEN &&
        strcmp(response + response_len - AT_RESPONSE_OK_LEN, AT_RESPONSE_OK) == 0)
        response[response_len - AT_RESPONSE_OK_LEN] = '\0';

    *res_param = strdup(response);
    if (!*res_param) {
        result = MCTD_RESULT_ERR_MEMORY;
        goto bail;
    }

bail:
    free(response);
    return result;
}

static int list_sms(char *req_param, char **res_param)
{
    int result;
    int fd;

    char *option = NULL;

    if (req_param)
        option = strtok(req_param, "\t");

    if (is_calling())
        return MCTD_RESULT_ERR_CALLING;

    fd = mctd_serial_open();
    if (fd < 0)
        return mctd_serial_convert_error(fd);

    result = list_sms_with_fd(fd, option, res_param);

    mctd_serial_close(fd);

    return result;
}

static int show_sms_with_fd(int fd, const char *sms_number,
                                                char **res_param)
{
    int result, response_len;
    char *response = NULL;
    char *command = "AT+CMGR";
    char buf[MCTD_AT_BUF_SIZE];

    if (!sms_number) {
        result = MCTD_RESULT_ERR_PARAM;
        goto bail;
    }

    snprintf(buf, MCTD_AT_BUF_SIZE, "%s=%s", command, sms_number);
    result = mctd_atcmd_send_and_echo_with_fd(fd, buf, &response, false, 1);
    if (result != MCTD_RESULT_OK)
        goto bail;

    if (!response) {
        result = MCTD_RESULT_ERR_PROCESS;
        goto bail;
    }

    /* cut response "OK" and line breaks before and after */
    response_len = strlen(response);
    if (response_len >= AT_RESPONSE_OK_LEN &&
        strcmp(response + response_len - AT_RESPONSE_OK_LEN, AT_RESPONSE_OK) == 0)
        response[response_len - AT_RESPONSE_OK_LEN] = '\0';

    *res_param = strdup(response);
    if (!*res_param) {
        result = MCTD_RESULT_ERR_MEMORY;
        goto bail;
    }

bail:
    free(response);
    return result;
}

static int show_sms(char *req_param, char **res_param)
{
    int fd;
    int result;

    char *sms_number = NULL;

    if (req_param)
        sms_number = strtok(req_param, "\t");
    if (!sms_number)
        return MCTD_RESULT_ERR_PARAM;

    if (is_calling())
        return MCTD_RESULT_ERR_CALLING;

    fd = mctd_serial_open();
    if (fd < 0)
        return mctd_serial_convert_error(fd);

    result = show_sms_with_fd(fd, sms_number, res_param);

    mctd_serial_close(fd);

    return result;
}

static int create_sms_with_fd(int fd, char *message, char *phone_number, char **res_param)
{
    int result;
    char *response = NULL;
    char *command = "AT+CMGW";
    char buf[MCTD_AT_BUF_SIZE];

    if (!message)
        return MCTD_RESULT_ERR_PARAM;

    if (!phone_number) {
        mctd_atcmd_send(fd, command, NULL, NULL, NULL);
    } else {
        snprintf(buf, MCTD_AT_BUF_SIZE, "%s=\"%s\"", command, phone_number);
        mctd_atcmd_send(fd, buf, NULL, NULL, NULL);
    }

    /* add ctrl+z */
    snprintf(buf, MCTD_AT_BUF_SIZE, "%s%c", message, 0x1a);
    result = mctd_atcmd_send(fd, buf, "+CMGW:", &response, NULL);

    if (result != MCTD_RESULT_OK)
        goto bail;

    if (!response) {
        result = MCTD_RESULT_ERR_PROCESS;
        goto bail;
    }

    cutCRandLF(response);
    *res_param = strdup(response);
    if (!*res_param) {
        result = MCTD_RESULT_ERR_MEMORY;
        goto bail;
    }


bail:
    free(response);
    return result;
}

static int create_sms(char *req_param, char **res_param)
{
    int fd;
    int result;

    char *message = NULL;
    char *phone_number = NULL;

    if (req_param)
        message = strtok(req_param, "\t");
    if (!message)
        return MCTD_RESULT_ERR_PARAM;

    phone_number = strtok(NULL, "\t");
    if (is_calling())
        return MCTD_RESULT_ERR_CALLING;

    fd = mctd_serial_open();
    if (fd < 0)
        return mctd_serial_convert_error(fd);

    result = create_sms_with_fd(fd, message, phone_number, res_param);

    mctd_serial_close(fd);

    return result;
}

static int send_sms_with_fd(int fd, const char *sms_number,
                           const char *phone_number)
{
    char *command = "AT+CMSS";
    char buf[MCTD_AT_BUF_SIZE];
    int timeout_sec = 10;

    if (!sms_number)
        return MCTD_RESULT_ERR_PARAM;

    if (!phone_number)
        snprintf(buf, MCTD_AT_BUF_SIZE, "%s=%s", command, sms_number);
    else
        snprintf(buf, MCTD_AT_BUF_SIZE, "%s=%s,\"%s\"", command, sms_number, phone_number);

    return mctd_atcmd_send(fd, buf, NULL, NULL, &timeout_sec);
}

static int send_sms(char *req_param, char **res_param)
{
    int fd;
    int result;

    char *sms_number = NULL;
    char *phone_number = NULL;

    UNUSED(res_param);

    if (req_param)
        sms_number = strtok(req_param, "\t");
    if (!sms_number)
        return MCTD_RESULT_ERR_PARAM;

    phone_number = strtok(NULL, "\t");

    if (is_calling())
        return MCTD_RESULT_ERR_CALLING;

    fd = mctd_serial_open();
    if (fd < 0)
        return mctd_serial_convert_error(fd);

    result = send_sms_with_fd(fd, sms_number, phone_number);

    mctd_serial_close(fd);

    return result;
}

static int delete_sms_with_fd(int fd, const char *sms_number)
{
    char *command = "AT+CMGD";
    char buf[MCTD_AT_BUF_SIZE];

    if (!sms_number)
        return MCTD_RESULT_ERR_PARAM;
    else
        snprintf(buf, MCTD_AT_BUF_SIZE, "%s=%s", command, sms_number);

    return mctd_atcmd_send(fd, buf, NULL, NULL, NULL);;
}

static int delete_sms(char *req_param, char **res_param)
{
    int fd;
    int result;

    char *sms_number = NULL;

    UNUSED(res_param);

    if (req_param)
        sms_number = strtok(req_param, "\t");
    if (!sms_number)
        return MCTD_RESULT_ERR_PARAM;

    if (is_calling())
        return MCTD_RESULT_ERR_CALLING;

    fd = mctd_serial_open();
    if (fd < 0)
        return mctd_serial_convert_error(fd);

    result = delete_sms_with_fd(fd, sms_number);

    mctd_serial_close(fd);

    return result;
}

static int set_sms_storage_with_fd(int fd, const char *storage, char **res_param)
{
    int result;
    char *response = NULL;
    char *command = "AT+CPMS";
    char buf[MCTD_AT_BUF_SIZE] = {0};

    if (!storage)
        snprintf(buf, MCTD_AT_BUF_SIZE, "%s?", command);
    else if (strcmp(storage, "sim") == 0 || strcmp(storage, "sm") == 0)
        snprintf(buf, MCTD_AT_BUF_SIZE, "%s=SM,SM,SM", command);
    else if (strcmp(storage, "me") == 0)
        snprintf(buf, MCTD_AT_BUF_SIZE, "%s=ME,ME,ME", command);
    else
        snprintf(buf, MCTD_AT_BUF_SIZE, "%s=%s", command, storage);

    result = mctd_atcmd_send(fd, buf, "+CPMS:", &response, NULL);

    if (result != MCTD_RESULT_OK)
        goto bail;

    if (!response) {
        result = MCTD_RESULT_ERR_PROCESS;
        goto bail;
    }

    cutCRandLF(response);
    *res_param = strdup(response);
    if (!*res_param) {
        result = MCTD_RESULT_ERR_MEMORY;
        goto bail;
    }

bail:
    free(response);
    return result;
}

static int set_sms_storage(char *req_param, char **res_param)
{
    int fd;
    int result;

    char *storage = NULL;

    if (req_param)
        storage = strtok(req_param, "\t");

    if (is_calling())
        return MCTD_RESULT_ERR_CALLING;

    fd = mctd_serial_open();
    if (fd < 0)
        return mctd_serial_convert_error(fd);

    result = set_sms_storage_with_fd(fd, storage, res_param);

    mctd_serial_close(fd);

    return result;
}

static int poweroff_and_fin(char *req_param, char **res_param)
{
    poweroff(req_param, res_param);
    return MCTD_RESULT_OK;
}

static const char *status_string()
{
    if (is_wvdial_running())
        return "dial";

    if (_no_sim_card)
        return "nosim";

    if (_registerd_operator_id)
        return "registered";

    return "not registered";
}

static float rsrq_to_db(int rsrq)
{
    return ((float)rsrq / 2.0) - 19.5;
}

static int put_rsrq_string(char *buf)
{
    if (_rsrq == 255)
        return sprintf(buf, "255: not known or not detectable");

    if (_rsrq > 34 || _rsrq < 0)
        return 0;

    if (_rsrq == 0)
        return sprintf(buf, " 0 [rsrq < -19.5 dB]");

    if (_rsrq == 34)
        return sprintf(buf, "34 [-3.0 dB < rsrq]");

    return sprintf(buf, "%2d [%2.1f dB <= rsrq < %2.1f dB]",
                        _rsrq, rsrq_to_db(_rsrq) - 0.5, rsrq_to_db(_rsrq));
}

static int rsrp_to_dbm(int rsrp)
{
    return rsrp - 140;
}

static int put_rsrp_string(char *buf)
{
    if (_rsrp == 255)
        return sprintf(buf, "255: not known or not detectable");

    if (_rsrp > 97 || _rsrp < 0)
        return 0;

    if (_rsrp == 0)
        return sprintf(buf, " 0 [rsrp < -140 dBm]");

    if (_rsrp == 97)
        return sprintf(buf, "97 [-44 dBm <= rsrp]");

    return sprintf(buf, "%2d [%3d dBm <= rsrp < %3d dBm]",
                        _rsrp, rsrp_to_dbm(_rsrp) - 1, rsrp_to_dbm(_rsrp));
}

static int status(char *req_param, char **res_param)
{
    int len = 0;
    char *p;

    UNUSED(req_param);

    *res_param = malloc(1024);
    if (!*res_param) {
        perror("malloc");
        return MCTD_RESULT_ERR_MEMORY;
    }

    p = *res_param;

    memset(p, 0, 1024);

    len += sprintf(&p[len], "    manufacturer: %s\n",
                                                _manufacturer?_manufacturer:"");
    len += sprintf(&p[len], "           model: %s\n", _model?_model:"");
    len += sprintf(&p[len], "        revision: %s\n", _revision?_revision:"");
    len += sprintf(&p[len], "     app revison: %s\n",
                                                _app_revision?_app_revision:"");
    len += sprintf(&p[len], "         port(s): %s\n", mctd_at_port_str() );
    len += sprintf(&p[len], "          status: %s\n", status_string());
    len += sprintf(&p[len], "    phone number: %s\n",
                                                _phone_number?_phone_number:"");
    len += sprintf(&p[len], "            imei: %s\n", _imei?_imei:"");
    len += sprintf(&p[len], "signale quality\n");

    len += sprintf(&p[len], "            rsrq: ");
    len += put_rsrq_string(&p[len]);
    len += sprintf(&p[len], "\n");

    len += sprintf(&p[len], "            rsrp: ");
    len += put_rsrp_string(&p[len]);
    len += sprintf(&p[len], "\n");

    len += sprintf(&p[len], "        sim slot: ");
    if (_sim_slot > 0)
        len += sprintf(&p[len], "%d", _sim_slot);
    len += sprintf(&p[len], "\n");

    len += sprintf(&p[len], "     operator id: %s\n",
                            _registerd_operator_id?_registerd_operator_id:"");

    return MCTD_RESULT_OK;
}

static void sync_datetime_with_fd(int fd)
{
    char *response = NULL;
    char datetime[21];

    if (get_datetime_with_fd(fd, &response) != MCTD_RESULT_OK) {
        log_info("get_datetime_with_fd failed.");
        return;
    }

    log_info("datetime[%s]", response);
    if (strncmp("70", response, 2) == 0) {
        /* no attach */
        log_info("no attach.");
        goto bail;
        return;
    }

    if (strlen(response) != 20) {
        log_info("datetime unexpected length.");
        return;
    }

    /* response is
       01234567890123456789
       yy/mm/dd,hh:mm:ss+tt */
    response[8] = ' ';
    response[17] = '\0';
    sprintf(datetime, "20%s", response);

    bin_execution(false, "hwclock", "--set", "--date", datetime, NULL);
    bin_execution(false, "hwclock", "-s", NULL);

bail:
    free(response);
}

static void sync_datetime()
{
    int fd;

    if (is_calling()) {
        log_info("calling.");
        return;
    }

    fd = mctd_serial_open();
    if (fd < 0) {
        log_info("open serial failed.");
        return;
    }

    sync_datetime_with_fd(fd);
    mctd_serial_close(fd);
}

static bool is_registered()
{
    char *operator_id = NULL;
    bool registered = false;

    if (is_wvdial_running())
        return true;

    if (!has_sim_card())
        return false;

    if (get_operator_id(&operator_id) == MCTD_RESULT_OK) {
        log_info("operator registered:%s", operator_id);
        free(_registerd_operator_id);
        _registerd_operator_id = malloc(strlen(operator_id) + 1);
        if (_registerd_operator_id) {
            strcpy(_registerd_operator_id, operator_id);
            _registerd_operator_id[strlen(operator_id)] = '\0';
        }
        get_signal_quality(NULL, NULL);
        mctd_atcmd_send_oneshot("AT^SMONI", NULL, NULL);
        registered = true;
    } else {
        free(_registerd_operator_id);
        _registerd_operator_id = NULL;
    }

    free(operator_id);
    return registered;
}

static int register_network()
{
    int result = MCTD_RESULT_ERR_NOT_REGISTERED;
    int i;
    struct timespec ts;

    ts.tv_sec = 5;
    ts.tv_nsec = 0;

    if (is_wvdial_running())
        return MCTD_RESULT_ERR_CALLING;

    if (!has_sim_card()) {
        log_info("no sim card.");
        return MCTD_RESULT_ERR_NO_SIM;
    }

    if (is_registered()) {
        result = MCTD_RESULT_OK_ALREADY;
        goto bail;
    }

    sleep(1);

    if (_operator_id) {
        if (set_operator_id(_operator_id) != MCTD_RESULT_OK) {
            if (set_cfun("0") != MCTD_RESULT_OK) {
                sleep(1);
                set_cfun("1");
                goto bail;
            }

            sleep(1);
            set_cfun("1");

            sleep(5);

            for (i = 0; i < 12; i++) {	/* 1 minite */
                if (set_operator_id(_operator_id) == MCTD_RESULT_OK)
                    break;

                /* thread cancellation points */
                nanosleep(&ts, NULL);
            }
        }
        sleep(1);
    }

    if (is_registered()) {
        result = MCTD_RESULT_OK;
        goto bail;
    }

    get_signal_quality(NULL, NULL);
    mctd_atcmd_send_oneshot("AT^SMONI", NULL, NULL);

bail:
    return result;
}

static void register_loop(bool first_time)
{
    int result = MCTD_RESULT_ERR_NOT_REGISTERED;
    int i = 0;
    int retry = first_time?3:1;

    if (is_wvdial_running())
        return;

    if (mctd_atcmd_send_oneshot("AT", NULL, NULL) != MCTD_RESULT_OK)
        return;

    while (result == MCTD_RESULT_ERR_NOT_REGISTERED) {
        result = register_network();
        switch(result) {
        case MCTD_RESULT_OK:
            sync_datetime();
            if (_auto_dial)
                dial(NULL, NULL);
            break;

        case MCTD_RESULT_ERR_NOT_REGISTERED:
            i++;
            if (i >= retry)
                return;

            break;

        case MCTD_RESULT_OK_ALREADY:
            if (first_time)
                sync_datetime();
            if (_auto_dial)
                dial(NULL, NULL);
            break;

        default:
            break;
        }
    }
}

static int registration(char *req_param, char **res_param)
{
    UNUSED(req_param);
    UNUSED(res_param);

    register_loop(false);

    return MCTD_RESULT_OK;
}

static int first_registration(char *req_param, char **res_param)
{
    UNUSED(req_param);
    UNUSED(res_param);

    register_loop(true);

    return MCTD_RESULT_OK;
}

static int ping_check(char *req_param, char **res_param)
{
    int result;

    if (!has_sim_card())
        return MCTD_RESULT_OK;

    if (!_ping_check)
        return MCTD_RESULT_OK;

    if (_auto_dial) {
        if (is_wvdial_running()) {
            if(bin_execution(false, "ping", "-s", "1", "-c", "2", "-w", "10",
                                                    _ping_ip_addr, NULL) == 0)
                result = MCTD_RESULT_OK;
            else
                result = MCTD_RESULT_ERR_NO_NETWORK;
        } else
            result = MCTD_RESULT_ERR_NO_NETWORK;
    } else {
        if (is_wvdial_running()) {
            if(bin_execution(false, "ping", "-s", "1", "-c", "2", "-w", "10",
                                                    _ping_ip_addr, NULL) == 0)
                result = MCTD_RESULT_OK;
            else
                result = MCTD_RESULT_ERR_NO_NETWORK;
        } else {
            if (_registerd_operator_id)
                result = MCTD_RESULT_OK;
            else
                result = MCTD_RESULT_ERR_NO_NETWORK;
        }
    }

    if (result == MCTD_RESULT_OK) {
        _failed_ping_check_count = 0;
        _network_restart_count = 0;
    } else {
        log_warn("failed ping check.");
        _failed_ping_check_count++;
    }

    if (_failed_ping_check_count > _ping_failed_threshold - 1) {
        /* first consecutive failed: network restart */
        if (_network_restart_count == 0) {
            set_cfun("0");
            sleep(1);
            set_cfun("1");
            _network_restart_count++;
        } else /* after 2 consecutive failed: modem restart */
            wwan_force_restart(req_param, res_param);

        _failed_ping_check_count = 0;
    }

    return result;
}

typedef struct {
  unsigned int cmd_id;
  int (*func)(char *, char **);
} CMD_T;

const CMD_T cmd_table[] = {
    { MCTD_CMD_SET_APN,            set_apn              },
    { MCTD_CMD_DIAL,               dial                 },
    { MCTD_CMD_HANGUP,             hangup               },
    { MCTD_CMD_GET_PHONE_NUMBER,   get_phone_number     },
    { MCTD_CMD_GET_SIGNAL_QUALITY, get_signal_quality   },
    { MCTD_CMD_WWAN_FORCE_RESTART, wwan_force_restart   },
    { MCTD_CMD_POWERON,            poweron              },
    { MCTD_CMD_POWEROFF,           poweroff             },
    { MCTD_CMD_SET_PSM,            set_psm              },
    { MCTD_CMD_SET_EDRX,           set_edrx             },
    { MCTD_CMD_SET_SUSPEND,        set_suspend          },
    { MCTD_CMD_SET_SLEEP,          set_sleep            },
    { MCTD_CMD_ACTIVATE,           activate             },
    { MCTD_CMD_DEACTIVATE,         deactivate           },
    { MCTD_CMD_SEND_AT,            send_at_command      },
    { MCTD_CMD_SEND_AT_ECHO,       send_at_command_echo },
    { MCTD_CMD_SET_SIM_SLOT,       set_sim_slot         },
    { MCTD_CMD_UPDATE_DATETIME,    update_datetime      },
    { MCTD_CMD_GET_DATETIME,       get_datetime         },
    { MCTD_CMD_POWEROFF_AND_FIN,   poweroff_and_fin     },
    { MCTD_CMD_FIN,                NULL                 },
    { MCTD_CMD_STATUS,             status               },
    { MCTD_CMD_LIST_SMS,           list_sms             },
    { MCTD_CMD_SHOW_SMS,           show_sms             },
    { MCTD_CMD_CREATE_SMS,         create_sms           },
    { MCTD_CMD_SEND_SMS,           send_sms             },
    { MCTD_CMD_DELETE_SMS,         delete_sms           },
    { MCTD_CMD_SET_SMS_STORAGE,    set_sms_storage      },
    { MCTD_CMD_REGISTRATION,       registration         },
    { MCTD_CMD_FIRST_REGISTRATION, first_registration   },
    { MCTD_CMD_POWER_SUPPLY_OFF,   power_supply_off     },
    { MCTD_CMD_PING_CHECK,         ping_check           },
    { MCTD_CMD_NOP,                NULL                 }
};

static void set_sleep_on_startup(const char *msec_str)
{
    long msec;
    char *end = NULL;
    char buf[MCTD_AT_BUF_SIZE];
    char *cp;

    if (!msec_str)
        return;

    if (check_param(msec_str, "disable")) {
        set_sleep("disable", NULL);
        return;
    }

    msec = strtol(msec_str, &end, 10);
    if (end && end[0] == '\0')
        end = NULL;

    if (end || msec < MCTD_SPOW_MIN_MSEC || msec > MCTD_SPOW_MAX_MSEC)
        return;

    sprintf(buf, "enable\t%s", msec_str);
    while((cp = strchr(buf, ',')))
        *cp = '\t';
    set_sleep(buf, NULL);
}

static void set_register_check_interval_on_startup(const char *interval_str)
{
    long interval;
    char *end = NULL;

    _check_register_interval_sec = 0;

    if (!interval_str)
        return;

    interval = strtol(interval_str, &end, 10);
    if (end && end[0] == '\0')
        end = NULL;

    if (end || interval < MCTD_CHECK_REGISTER_INTERVAL_SEC_MIN ||
        interval > MCTD_CHECK_REGISTER_INTERVAL_SEC_MAX) {
        log_warn("interval[%d]", interval);
        return;
    }

    _check_register_interval_sec = interval;
}

static void set_ping_failed_threshold_on_startup(const char *count_str)
{
    long count;
    char *end = NULL;

    _ping_failed_threshold = 1;

    if (!count_str)
        return;

    count = strtol(count_str, &end,  10);
    if (end && end[0] == '\0')
        end = NULL;

    if (end || count < 1) {
        log_warn("count[%d]", count);
        return;
    }

    _ping_failed_threshold = count;
}

static void set_ping_ip_addr_on_startup(const char *ip_addr)
{
    if (ip_addr) {
        free(_ping_ip_addr);
        _ping_ip_addr = strdup(ip_addr);
    }
}

static void startup(const char *conf_file_path)
{
    MCTD_TYPE_AND_VALUE_ARRAY_T params;
    char *apn = NULL;
    char *auth_prot = NULL;
    char *username = NULL;
    char *password = NULL;
    char *operator_id = NULL;
    char *fix_profile = NULL;
    char *sleep = NULL;
    char *suspend = NULL;
    char *psm = NULL;
    char *auto_attach = NULL;
    char *edrx = NULL;
    char *auto_dial = NULL;
    char *multiplex = NULL;
    char *create_wvdial_conf = NULL;
    char *register_check_interval = NULL;
    char *ping_check = NULL;
    char *ping_ip_addr = NULL;
    char *ping_failed_threshold = NULL;
    char buf[MCTD_AT_BUF_SIZE];
    char *cp;

    if (mctd_atcmd_send_oneshot("AT", NULL, NULL) != MCTD_RESULT_OK)
        return;

    if (!conf_file_path)
        return;

    if (!mctd_startup_parse(conf_file_path, &params))
        return;

    apn         = mctd_startup_get_value(&params, "apn");
    auth_prot   = mctd_startup_get_value(&params, "auth_prot");
    username    = mctd_startup_get_value(&params, "username");
    password    = mctd_startup_get_value(&params, "password");
    operator_id = mctd_startup_get_value(&params, "operator_id");
    fix_profile = mctd_startup_get_value(&params, "fix_profile");
    sleep       = mctd_startup_get_value(&params, "sleep");
    suspend     = mctd_startup_get_value(&params, "suspend");
    psm         = mctd_startup_get_value(&params, "psm");
    edrx        = mctd_startup_get_value(&params, "edrx");
    auto_dial   = mctd_startup_get_value(&params, "auto_dial");
    multiplex   = mctd_startup_get_value(&params, "multiplex");
    auto_attach = mctd_startup_get_value(&params, "auto_attach");
    create_wvdial_conf = mctd_startup_get_value(&params, "create_wvdial_conf");
    register_check_interval = mctd_startup_get_value(&params,
                                                    "register_check_interval");
    ping_check  = mctd_startup_get_value(&params, "ping_check");
    ping_ip_addr = mctd_startup_get_value(&params, "ping_ip_addr");
    ping_failed_threshold = mctd_startup_get_value(&params, "ping_failed_threshold");

    if (apn && apn[0] == '\0')
        apn = NULL;

    if (operator_id && operator_id[0] == '\0')
        operator_id = NULL;

    if (fix_profile && fix_profile[0] == '\0')
        fix_profile = NULL;

    if (sleep && sleep[0] == '\0')
        sleep = NULL;

    if (suspend && suspend[0] == '\0')
        suspend = NULL;

    if (psm && psm[0] == '\0')
        psm = NULL;

    if (edrx && edrx[0] == '\0')
        edrx = NULL;

    if (auto_attach && auto_attach[0] == '\0')
        auto_attach = NULL;

    if (ping_ip_addr && ping_ip_addr[0] == '\0')
        ping_ip_addr = NULL;

    if (ping_failed_threshold && ping_failed_threshold[0] == '\0')
        ping_failed_threshold = NULL;

    set_register_check_interval_on_startup(register_check_interval);

    _multiplex = MUX_STATUS_DISABLE;
    strcpy(ppp_port, SERIAL_TTY);
    strcpy(at_port, SERIAL_TTY);
    if (multiplex) {
        if (check_param(multiplex, "enable")) {
            _multiplex = MUX_STATUS_ENABLE;
            if (mctd_mux_enable()) {
                log_warn("failed apply multiplex.");
                _multiplex = MUX_STATUS_DISABLE;
            } else {
                strcpy(ppp_port, MUX_PPP_TTY);
            }
        } else if(check_param(multiplex, "auto")) {
            _multiplex = MUX_STATUS_AUTO;
            if (mctd_mux_enable()) {
                _multiplex = MUX_STATUS_DISABLE;
            } else {
                strcpy(ppp_port, MUX_PPP_TTY);
            }
        }
    }

    if (create_wvdial_conf) {
        if (check_param(create_wvdial_conf, "true")) {
            if (apn)
                mctd_create_wvdial_conf(apn, username, password);
        }
    }

    if (apn && auth_prot) {
        set_apn_and_auth(apn, auth_prot, username, password);
    } else if (apn) {
        set_apn(apn, NULL);
    } else if (auth_prot) {
        set_auth(auth_prot, username, password);
    }

    if (auto_attach) {
        if (check_param(auto_attach, "disable"))
            set_auto_attach(false);
        else if (check_param(auto_attach, "enable"))
            set_auto_attach(true);
    }

    if (operator_id) {
        _operator_id = malloc(strlen(operator_id) + 1);
        if (!_operator_id)
            log_err("malloc failed (_operator_id).");
        strcpy(_operator_id, operator_id);
    } else
        _operator_id = NULL;

    if (fix_profile) {
        if (check_param(fix_profile, "none"))
            set_fix_profile(NULL);
        else
            set_fix_profile(fix_profile);
    }

    set_sleep_on_startup(sleep);

    if (suspend) {
        if (check_param(suspend, "disable"))
            set_suspend("disable", NULL);
        else if (check_param(suspend, "enable")) {
            set_suspend("enable", NULL);
            activate(NULL, NULL);
        }
    }

    if (psm) {
        if (check_param(psm, "disable"))
            set_psm("disable", NULL);
        else {
            sprintf(buf, "enable\t%s", psm);
            while((cp = strchr(buf, ',')))
                *cp = '\t';
            set_psm(buf, NULL);
        }
    }

    if (edrx) {
        if (check_param(edrx, "disable"))
            set_edrx("disable", NULL);
        else {
            sprintf(buf, "enable\t%s", edrx);
            while((cp = strchr(buf, ',')))
                *cp = '\t';
            set_edrx(buf, NULL);
        }
    }

    _auto_dial = false;
    if (auto_dial) {
        if (check_param(auto_dial, "true")) {
            _auto_dial = true;
        }
    }

    set_ping_ip_addr_on_startup(ping_ip_addr);

    set_ping_failed_threshold_on_startup(ping_failed_threshold);

    _ping_check = false;
    if (ping_check) {
        if (check_param(ping_check, "true")) {
            if (!_ping_ip_addr || !_ping_failed_threshold) {
                log_warn("not set ping_ip_addr or ping_failed_threshold");
            } else
                _ping_check = true;
        }
    }

    mctd_startup_free(&params);
}

static bool put_model_string(char *str, char **to)
{
    free(*to);
    *to = malloc(strlen(str) + 1);
    if (!*to)
        return false;;

    strcpy(*to, str);
    (*to)[strlen(str)] = '\0';

    return true;
}

static int model_string_with_fd(int fd)
{
    int result;
    char *ati1 = NULL;
    char *p;

    /* need echo */
    mctd_atcmd_send(fd, "ATE1", NULL, NULL, NULL);

    result = mctd_atcmd_send_and_echo_with_fd(fd, "ATI1", &ati1, true, 30);
    if (result != MCTD_RESULT_OK)
        return result;

    if (!ati1)
        return MCTD_RESULT_ERR_PROCESS;

    result = MCTD_RESULT_ERR_PROCESS;

    p = strtok(ati1, "\r\n");
    if (!p)
        goto bail;

    p = strtok(NULL, "\r\n");
    if (!p)
        goto bail;

    if (!put_model_string(p, &_manufacturer)) {
        result = MCTD_RESULT_ERR_MEMORY;
        goto bail;
    }
    p = strtok(NULL, "\r\n");
    if (!p)
        goto bail;

    if (!put_model_string(p, &_model)) {
        result = MCTD_RESULT_ERR_MEMORY;
        goto bail;
    }

    p = strtok(NULL, "\r\n");
    if (!p)
        goto bail;

    if (!put_model_string(p, &_revision)) {
        result = MCTD_RESULT_ERR_MEMORY;
        goto bail;
    }

    p = strtok(NULL, "\r\n");
    if (!p)
        goto bail;

    if (!put_model_string(p, &_app_revision)) {
        result = MCTD_RESULT_ERR_MEMORY;
        goto bail;
    }

    result = MCTD_RESULT_OK;

bail:
    free(ati1);
    return result;
}

static void startup_modem()
{
    int fd;
    int slot;

    fd = mctd_serial_open();
    if (fd < 0) {
        log_warn("open serial failed.");
        return;
    }

    set_errot_format_with_fd(fd);
    model_string_with_fd(fd);
    sim_slot_with_fd(fd, &slot);
    imei_with_fd(fd, NULL);
    set_sms_mode_with_fd(fd, "1");

    if (!is_match_cfun_with_fd(fd, '1'))
        set_cfun_with_fd(fd, "1");

    set_scfg_gpio_mode_ring0_std_with_fd(fd);
    set_scfg_gpio_mode_asc1_rsv_with_fd(fd);
    set_scfg_urc_ringline_asc0_with_fd(fd);
    set_scfg_urc_ringline_activetime_1s_with_fd(fd);
    set_sgpicfg_logic_ring0_low_with_fd(fd);
    set_sgpicfg_port_wakein_gpio25_with_fd(fd);

    mctd_serial_close(fd);

    get_phone_number(NULL, NULL);
}

bool mctd_modem_control_has_cmd_id(unsigned int cmd_id)
{
    int i;
    bool result;

    for (i = 0; ; i++) {
        if (cmd_table[i].cmd_id == MCTD_CMD_NOP) {
            result = false;
            break;
        }
        if (cmd_table[i].cmd_id == cmd_id) {
            result = true;
            break;
        }
    }

    return result;
}

int mctd_modem_control_exec_app_cmd(unsigned int cmd_id, char *req_param, char **res_param)
{
    int i;

    for (i = 0; ; i++) {
        if (cmd_table[i].cmd_id == MCTD_CMD_NOP) {
            break;
        }
        if (cmd_table[i].cmd_id == cmd_id) {
            return cmd_table[i].func(req_param, res_param);
        }
    }

    return MCTD_RESULT_ERR_CMD_NOT_FOUND;
}

bool mctd_modem_control_is_suspend()
{
    return _is_suspend;
}

int mctd_modem_control_get_check_register_interval_sec()
{
    return _check_register_interval_sec;
}

void mctd_modem_control_finalize()
{
    free(_operator_id);
    free(_manufacturer);
    free(_model);
    free(_revision);
    free(_app_revision);
    free(_phone_number);
    free(_imei);
    free(_registerd_operator_id);
    free(_ping_ip_addr);
}
